/**
 * Mupen64 - memory.h
 * Copyright (C) 2002 Hacktarux
 *
 * Mupen64 homepage: http://mupen64.emulation64.com
 * email address: hacktarux@yahoo.fr
 * 
 * If you want to contribute to the project please contact
 * me first (maybe someone is already making what you are
 * planning to do).
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
**/

#ifndef MEMORY_H
#define MEMORY_H

#include "tlb.h"

int init_memory();
void free_memory();
#define read_word_in_memory() readmem[address>>16]()
#define read_byte_in_memory() readmemb[address>>16]()
#define read_hword_in_memory() readmemh[address>>16]()
#define read_dword_in_memory() readmemd[address>>16]()
#define write_word_in_memory() writemem[address>>16]()
#define write_byte_in_memory() writememb[address >>16]()
#define write_hword_in_memory() writememh[address >>16]()
#define write_dword_in_memory() writememd[address >>16]()
extern unsigned long SP_DMEM[0x1000/4];
extern unsigned char *SP_DMEMb;
extern unsigned long SP_IMEM[0x1000/4];
extern unsigned long PIF_RAM[0x40/4];
extern unsigned char *PIF_RAMb;
extern unsigned long rdram[0x400000/4];
extern unsigned long address, word;
extern unsigned char byte;
extern unsigned short hword;
extern unsigned long long int dword, *rdword;

extern void (*readmem[0xFFFF])();
extern void (*readmemb[0xFFFF])();
extern void (*readmemh[0xFFFF])();
extern void (*readmemd[0xFFFF])();
extern void (*writemem[0xFFFF])();
extern void (*writememb[0xFFFF])();
extern void (*writememd[0xFFFF])();
extern void (*writememh[0xFFFF])();

typedef struct _mips_register
{
   char init_length;
   char init_mode;
   char ebus_test_mode;
   char RDRAM_reg_mode;
   unsigned long w_mi_init_mode_reg;
   unsigned long mi_init_mode_reg;
   unsigned long mi_version_reg;
   unsigned long mi_intr_reg;
   unsigned long mi_intr_mask_reg;
   unsigned long w_mi_intr_mask_reg;
   char SP_intr_mask;
   char SI_intr_mask;
   char AI_intr_mask;
   char VI_intr_mask;
   char PI_intr_mask;
   char DP_intr_mask;
} mips_register;

typedef struct _PI_register
{
   unsigned long pi_dram_addr_reg;
   unsigned long pi_cart_addr_reg;
   unsigned long pi_wr_len_reg;
   unsigned long read_pi_status_reg;
} PI_register;

typedef struct _SP_register
{
   unsigned long sp_mem_addr_reg;
   unsigned long sp_dram_addr_reg;
   unsigned long sp_rd_len_reg;
   unsigned long sp_wr_len_reg;
   unsigned long w_sp_status_reg;
   unsigned long sp_status_reg;
   char halt;
   char broke;
   char single_step;
   char intr_break;
   char signal0;
   char signal1;
   char signal2;
   char signal3;
   char signal4;
   char signal5;
   char signal6;
   char signal7;
   unsigned long sp_dma_full_reg;
   unsigned long sp_dma_busy_reg;
   unsigned long sp_semaphore_reg;
} SP_register;

typedef struct _DPC_register
{
   unsigned long dpc_start;
   unsigned long dpc_end;
   unsigned long dpc_current;
   unsigned long dpc_clock;
   unsigned long dpc_bufbusy;
   unsigned long dpc_pipebusy;
   unsigned long dpc_tmem;
   unsigned long w_dpc_status;
   unsigned long dpc_status;
   char xbus_dmem_dma;
} DPC_register;

typedef struct _SI_register
{
   unsigned long si_dram_addr;
   unsigned long si_pif_addr_rd64b;
   unsigned long si_pif_addr_wr64b;
   unsigned long si_status;
} SI_register;

typedef struct _AI_register
{
   unsigned long ai_dram_addr;
   unsigned long ai_len;
   unsigned long ai_control;
   unsigned long ai_status;
   unsigned long ai_dacrate;
   unsigned long ai_bitrate;
} AI_register;

typedef struct _VI_register
{
   unsigned long vi_status;
   unsigned long vi_origin;
   unsigned long vi_width;
   unsigned long vi_v_intr;
   long vi_current;
   unsigned long vi_burst;
   unsigned long vi_v_sync;
   unsigned long vi_h_sync;
   unsigned long vi_leap;
   unsigned long vi_h_start;
   unsigned long vi_v_start;
   unsigned long vi_v_burst;
   unsigned long vi_x_scale;
   unsigned long vi_y_scale;
} VI_register;

typedef struct _RI_register
{
   unsigned long ri_mode;
   unsigned long ri_config;
   unsigned long ri_current_load;
   unsigned long ri_select;
   unsigned long ri_refresh;
} RI_register;

typedef struct _RDRAM_register
{
   unsigned long rdram_mode;
} RDRAM_register;

extern PI_register pi_register;
extern mips_register MI_register;
extern SP_register sp_register;
extern SI_register si_register;
extern VI_register vi_register;
extern unsigned long RSP_PC_REG;
extern RI_register ri_register;
extern AI_register ai_register;
extern DPC_register dpc_register;

#ifndef _BIG_ENDIAN
#define sl(mot) \
( \
((mot & 0x000000FF) << 24) | \
((mot & 0x0000FF00) <<  8) | \
((mot & 0x00FF0000) >>  8) | \
((mot & 0xFF000000) >> 24) \
)

#define S8 3
#define S16 2
#define Sh16 1

#else

#define sl(mot) mot
#define S8 0
#define S16 0
#define Sh16 0

#endif

void read_nothing();
void read_nothingh();
void read_nomem();
void read_nomemb();
void read_nomemh();
void read_nomemd();
void read_rdram();
void read_rdramb();
void read_rdramh();
void read_rdramd();
void read_rdramreg();
void read_rsp_mem();
void read_rsp_memb();
void read_rsp_reg();
void read_rsp();
void read_mi();
void read_vi();
void read_ai();
void read_pi();
void read_ri();
void read_si();
void read_rom();
void read_pif();
void read_pifb();

void write_nothing();
void write_nomem();
void write_nomemb();
void write_nomemd();
void write_nomemh();
void write_rdram();
void write_rdramb();
void write_rdramh();
void write_rdramd();
void write_rdram_int();
void write_rdram_intb();
void write_rdram_inth();
void write_rdram_intd();
void write_actual_block();
void write_actual_blockb();
void write_actual_blockd();
void write_actual_blockh();
void write_rdramreg();
void write_rsp_mem();
void write_rsp_memb();
void write_rsp_reg();
void write_rsp();
void write_mi();
void write_vi();
void write_ai();
void write_ri();
void write_si();
void write_pi();
void write_pif();

void update_SP();
void update_DPC();

#endif
